/*    
    th_dnh_archiver
    Copyright (C) 2010-2011 Lymia <lymiahugs@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package lymia.th_dnh.archiver;

import java.io.File;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.io.RandomAccessFile;
import java.util.Arrays;

class ArchiveAccess {
	static final byte[] MAGIC_ARCHIVE = "PACK_FILE\u0000".getBytes(); 
	static final byte[] MAGIC_COMPRESS = "COMPRESS_ZIP\u0000".getBytes();
	
	private final RandomAccessFile file;

	public ArchiveAccess(File file) throws FileNotFoundException {
		this.file = new RandomAccessFile(file, "rw");
	}
	
	public RandomAccessFile getRandomAccessFile() {
		return file;
	}
	
	public void expand(long length) throws IOException {
		file.setLength(Math.max(file.getFilePointer()+length,file.length()));
	}
	
	public void close() throws IOException {
		file.close();
	}
	public long getFilePointer() throws IOException {
		return file.getFilePointer();
	}
	public long length() throws IOException {
		return file.length();
	}
	public void seek(long pos) throws IOException {
		file.seek(pos);
	}
	public void setLength(long newLength) throws IOException {
		file.setLength(newLength);
	}
	public void reset() throws IOException {
		file.seek(0);
	}

	public void checkMagicNumber() throws DanmakufuException, IOException {
		byte[] buffer = new byte[MAGIC_ARCHIVE.length];
		file.readFully(buffer);
		
		if(!Arrays.equals(buffer, MAGIC_ARCHIVE)) throw new DanmakufuException("bad magic number");
	}
	public boolean isZipMarker() throws IOException {
		long resetMark = file.getFilePointer();
		
		byte[] buffer = new byte[MAGIC_COMPRESS.length];
		file.readFully(buffer);

		if(!Arrays.equals(buffer, MAGIC_COMPRESS)) {
			file.seek(resetMark);
			return false;
		} else return true;
	}
	public long readInt() throws IOException {
		return (long)(file.read())    |
			   (long)(file.read())<<8 |
			   (long)(file.read())<<16|
			   (long)(file.read())<<24;
	}
	public String readString() throws IOException {
		long length = readInt();
		if((int)length!=length) throw new IOException("string too long");
		byte[] buffer = new byte[(int) (length-1)];
		file.readFully(buffer);
		file.read();
		
		return new String(buffer);
	}
	public byte[] read(int length) throws IOException {
		byte[] data = new byte[length];
		file.readFully(data);
		return data;
	}
	
	public void writeMagicNumber() throws IOException {
		file.write(MAGIC_ARCHIVE);
	}
	public void writeZipMarker() throws IOException {
		file.write(MAGIC_COMPRESS);
	}
	public void writeInt(long i) throws IOException {
		file.write((int)((i    )&0xFF));
		file.write((int)((i>>8 )&0xFF));
		file.write((int)((i>>16)&0xFF));
		file.write((int)((i>>24)&0xFF));
	}
	public void writeString(String s) throws IOException {
		byte[] string = s.getBytes();
		writeInt(string.length+1);
		file.write(string);
		file.write(0);
	}
	public void write(byte[] data) throws IOException {
		file.write(data);
	}
}
