/*    
    th_dnh_archiver
    Copyright (C) 2010-2011 Lymia <lymiahugs@gmail.com>

    This program is free software: you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation, either version 3 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program.  If not, see <http://www.gnu.org/licenses/>.
*/
package lymia.th_dnh.archiver;

import java.io.ByteArrayOutputStream;
import java.io.DataInputStream;
import java.io.EOFException;
import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.FileOutputStream;
import java.io.IOException;
import java.io.InputStream;
import java.io.OutputStream;
import java.io.RandomAccessFile;
import java.util.HashMap;
import java.util.WeakHashMap;
import java.util.zip.DeflaterOutputStream;
import java.util.zip.InflaterInputStream;

public final class DanmakufuArchive {
	private static final int TRANSFER_BUFFER_SIZE = 0x400000; //4MB
	
	private File archivePath;
	private ArchiveAccess access;
	
	private HashMap<String,Entry> entries = new HashMap<String,Entry>();
	private WeakHashMap<String,byte[]> entryContentBuffer = new WeakHashMap<String,byte[]>();
	
	private byte[] buffer = new byte[TRANSFER_BUFFER_SIZE];
	
	public DanmakufuArchive(File archive) throws IOException {
		archivePath = archive;
		access = new ArchiveAccess(archive);
	}
	
	public void readEntryList() throws IOException, DanmakufuException {
		HashMap<String,Entry> entries = new HashMap<String,Entry>();
		
		try {
			access.reset();
			access.checkMagicNumber();
			
			long entriesTemp = access.readInt();
			int entryCount = (int)entriesTemp;
			if(entriesTemp!=entryCount) throw new DanmakufuException("too many files");
			
			long fileLength = access.length();
			
			for(int i=0;i<entryCount;i++) {
				String name = access.readString();
				if(entries.containsKey(name)) throw new DanmakufuException("duplicate name");
				
				long offset = access.readInt();
				long length = access.readInt();
				
				if(offset+length>fileLength) throw new DanmakufuException("bad location");
				
				entries.put(name, new Entry(offset,length));
			}
		} catch(EOFException e) {
			throw new DanmakufuException("unexpected end of file", e);
		}
		
		this.entries = entries;
		this.entryContentBuffer = new WeakHashMap<String,byte[]>();
	}
	
	public void writeArchive() throws IOException {
		writeArchive(archivePath, true);
	}
	public void writeArchive(File f) throws IOException {
		writeArchive(f, true);
	}
	public void writeArchive(boolean compress) throws IOException {
		writeArchive(archivePath, compress);
	}
	public void writeArchive(File target, boolean compress) throws IOException {
		File tempFile = File.createTempFile("danmakufu-archiver-", ".dat", target.getParentFile());
		ArchiveAccess output = new ArchiveAccess(tempFile);
		
		try {
			HashMap<String,Long> entryOffsetAddresses = new HashMap<String,Long>();
			
			output.writeMagicNumber();
			output.writeInt(entries.size());
			for(String name:entries.keySet()) {
				Entry e = entries.get(name);
				
				output.writeString(name);
				entryOffsetAddresses.put(name, output.getFilePointer());
				output.writeInt(0);
				output.writeInt(e.length);
			}
	
			for(String name:entries.keySet()) {
				Entry e = entries.get(name);
				OutputStream o = randomAccessFileOutput(output.getRandomAccessFile());
	
				long offset = output.getFilePointer();
				
				ByteArrayOutputStream b = null;
				DeflaterOutputStream z = null;
				if(compress) {
					b = new ByteArrayOutputStream();
					z = new DeflaterOutputStream(o);
					output.writeZipMarker();
					output.writeInt(e.length);
				} else output.expand(e.length);
				
				copyFromEntry(e,compress?z:o);
				
				if(compress) {
					z.finish();
					output.expand(b.size());
					output.write(b.toByteArray());
				}
				
				long mark = output.getFilePointer();
				output.seek(entryOffsetAddresses.get(name));
				output.writeInt(offset);
				if(compress) output.writeInt(b.size());
				output.seek(mark);
			}
		} finally {
			output.close();
		}
			
		if(target.equals(archivePath)) access.close();
		if(target.exists()) if(!target.delete()) throw new IOException("failed to delete old file");
		while(target.exists()) try {Thread.sleep(1);} catch (InterruptedException e1) {}
		if(!tempFile.renameTo(target)) throw new IOException("failed to move temp file");
		if(target.equals(archivePath)) try {
			access = new ArchiveAccess(archivePath);
			readEntryList();
		} catch (DanmakufuException e) {
			throw new RuntimeException("unexpected error: newly written archive is invalid", e);
		}
	}

	public String[] getEntries() {
		String[] entries = new String[this.entries.size()];
		this.entries.keySet().toArray(entries);
		return entries;
	}
	
	public byte[] getEntryData(String entry) throws IOException, DanmakufuException {
		byte[] buffered = entryContentBuffer.get(entry);
		if(buffered!=null) return buffered;
		
		if(!entries.containsKey(entry)) return null;;
		
		Entry e = entries.get(entry);
		if(e.length>Integer.MAX_VALUE)
			throw new DanmakufuException("target entry is larger than max array size");
		
		ByteArrayOutputStream out = new ByteArrayOutputStream((int)e.length);
		copyFromEntry(e,out);
		return out.toByteArray();
	}
	public void extract(File targetDirectory) throws IOException {
		if(targetDirectory.exists()&&!targetDirectory.isDirectory())
			throw new IOException("target file is not directory");
		
		targetDirectory.mkdirs();
		
		for(String entry:entries.keySet()) 
			writeEntryToFile(entry, new File(targetDirectory,entry));
	}
	public void writeEntryToFile(String entry, File file) throws IOException {
		FileOutputStream out = new FileOutputStream(file);
		
		if(!entries.containsKey(entry)) throw new FileNotFoundException("no such entry");
		
		byte[] buffered = entryContentBuffer.get(entry);
		if(buffered!=null) out.write(buffered);
		else copyFromEntry(entries.get(entry),out);
		
		out.flush();
		out.close();
	}
	
	public void addEntryFromArray(String entry, byte[] data) {
		entries.put(entry,new Entry(data.clone()));
		entryContentBuffer.remove(entry);
	}
	public void addEntryFromPath(String entry, File file) {
		if(file.isFile()) {
			entries.put(entry,new Entry(file));
			entryContentBuffer.remove(entry);
		} else if(file.isDirectory()) {
			for(File f:file.listFiles())
				addEntryFromPath(f.getName(),f);
		}
	}
	
	public void removeEntry(String entry) {
		entries.remove(entry);
		entryContentBuffer.remove(entry);
	}
	
	private void copyFromEntry(Entry e, OutputStream dest) throws IOException {
		switch(e.location) {
			case ARCHIVE:
				access.seek(e.offset);
				if(access.isZipMarker()) {
					long size = access.readInt();
					InflaterInputStream in = 
						new InflaterInputStream(randomAccessFileInput(access.getRandomAccessFile(),
												e.length-ArchiveAccess.MAGIC_COMPRESS.length-4));
					copyFromStream(size, in, dest);
				} else copyFromRandomAccessFile(e.offset, e.length, access.getRandomAccessFile(), dest);
				break;
			case DISK:
				File f = e.file;
				FileInputStream s = new FileInputStream(f);
				try {
					copyFromStream(f.length(), s, dest);
				} finally {
					s.close();
				}
				break;
			case MEMORY:
				dest.write(e.data);
				break;
		}
	}
	
	private static InputStream randomAccessFileInput(final RandomAccessFile r, final long size) {
		return new InputStream() {
			long count = size;
			public int read() throws IOException {
				if(count==0) return -1;
				count--;
				return r.read();
			}
		};
	}
	private static OutputStream randomAccessFileOutput(final RandomAccessFile r) {
		return new OutputStream() {
			public void write(int b) throws IOException {
				r.write(b);
			}
		};
	}
	
	private void copyFromRandomAccessFile(long offset, long length, RandomAccessFile src, OutputStream dest) throws IOException {
		if(length==0) return;
		
		src.seek(offset);
		while(length>=0) {
			long size = Math.min(length, TRANSFER_BUFFER_SIZE);
			src.readFully(buffer,0,(int)size);
			dest.write(buffer,0,(int)size);
			length -= TRANSFER_BUFFER_SIZE;
		}
	}
	
	private void copyFromStream(long length, InputStream src, OutputStream dest) throws IOException {
		if(length==0) return;
		DataInputStream in = new DataInputStream(src);
		while(length>=0) {
			long size = Math.min(length, TRANSFER_BUFFER_SIZE);
			in.readFully(buffer,0,(int)size);
			dest.write(buffer,0,(int)size);
			length -= TRANSFER_BUFFER_SIZE;
		}
	}
	
	private static enum Location {
		ARCHIVE,
		DISK,
		MEMORY
	}
	private static class Entry {
		public final Location location;
		public final long length;
		
		public final long offset;
		public final byte[] data;
		public final File file;
		
		public Entry(long offset, long length) {
			location = Location.ARCHIVE;
			this.offset = offset;
			this.length = length;
			
			data = null;
			file = null;
		}
		public Entry(byte[] data) {
			location = Location.MEMORY;
			this.data = null;
			length = data.length;
			
			offset = -1;
			file = null;
		}
		public Entry(File file) {
			location = Location.DISK;
			this.file = file;
			length = file.length();
			
			offset = -1;
			data = null;
		}
	}
}
